/*
** Copyright (c) 2002 D. Richard Hipp
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** General Public License for more details.
** 
** You should have received a copy of the GNU General Public
** License along with this library; if not, write to the
** Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA  02111-1307, USA.
**
** Author contact information:
**   drh@hwaci.com
**   http://www.hwaci.com/drh/
**
*******************************************************************************
**
** This file contains code used to generate convert plain text into
** similarly formated HTML.
*/
#include "config.h"
#include "format.h"
#include <time.h>


/*
** Return the number digits at the beginning of the string z.
*/
static int ndigit(const char *z){
  int i = 0;
  while( isdigit(*z) ){ i++; z++; }
  return i;
}

/*
** Check to see if *z contains nothing but spaces up to the next
** newline.  If so, return the number of spaces plus one for the
** newline characters.  If not, return 0.
**
** If two or more blank lines occur in a row, go ahead and return
** a number of characters sufficient to cover them all.
*/
static int is_blank_line(const char *z){
  int i = 0;
  int r = 0;
  while( isspace(z[i]) ){
    if( z[i]=='\n' ){ r = i+1; }
    i++;
  }
  return r;
}

/*
** Return TRUE if *z points to the terminator for a word.  Words
** are terminated by whitespace or end of input or any of the
** characters in zEnd.
*/
static int is_eow(const char *z, const char *zEnd){
  if( zEnd==0 ) zEnd = ".,:;?!)\"'";
  while( *z!=0 && !isspace(*z) ){
    int i;
    for(i=0; zEnd[i]; i++){ if( *z==zEnd[i] ) break; }
    if( zEnd[i]==0 ) return 0;
    z++;
  }
  return 1;
}

/*
** Check to see if *z points to the beginning of a Wiki page name.
** If it does, return the number of characters in that name.  If not,
** return 0.
**
** A Wiki page name contains only alphabetic characters.  The first
** letter must be capital and there must be at least one other capital
** letter in the word.  And every capital leter must be followed by
** one or more lower-case letters.
**
** There are two versions of this routine.  is_wiki_name() tolerates
** no underscore characters in the name .  is_wiki_name_() allows the
** last character of the name to be an underscore.  Both versions of
** the routine call a common core.
*/
static int _is_wiki_name_core(const char *z, int allowUnderscore){
  int i;
  int nCap = 0;
  if( !isupper(z[0]) ) return 0;
  for(i=0; z[i]; i++){
    if( isupper(z[i]) ){
      if( !islower(z[i+1]) ) return 0;
      nCap++;
    }else if( !islower(z[i]) ){
      break;
    }
  }
  if( allowUnderscore && (z[i]=='_' || z[i]=='*') ) i++;
  return nCap>=2 && is_eow(&z[i],0) ? i : 0;
}
int is_wiki_name(const char *z){ return _is_wiki_name_core(z, 0); }
int is_wiki_name_(const char *z){ return _is_wiki_name_core(z, 1); }

#if 0 /* NOT USED */
/*
** If *z points to a word or phrase that should be bold, return
** the number of characters in the word or phrase.  Otherwise return 0.
**
** A word should be bold if it begins and ends with and underscore.
*/
static int is_bold_word(const char *z){
  int i, b = 0;
  if( z[0]!='_' ) return 0;
  for(i=1; z[i] && !isspace(z[i]); i++){ if( z[i]=='_' ) b = i; }
  if( b>1 && is_eow(&z[b+1],0) ) return b+1;
  return 0;
}
#endif

/*
** If *z points to text of the form:  "N: " where N is a sequence
** of one or more digits,
** then *z points to the start of an enumeration item.  Return the
** number of digits in N plus 1.  (Minimum return value is 2.)
** If *z does not point to such text, return 0.
*/
int is_enum(const char *z){
  int i;
  for(i=0; isdigit(z[i]); i++){}
  if( i<1 ) return 0;
  if( z[i]!=':' ) return 0;
  if( z[i+1]!=' ' && z[i+1]!='\t' ) return 0;
  return i+1;
}

/*
** If *z points to text of the form:  "*: " 
** then *z points to the start of a bullet.  Return 2.
** If *z does not point to such text, return 0.
*/
int is_bullet(const char *z){
  if( z[0]!='*' ) return 0;
  if( z[1]!=':' ) return 0;
  if( z[2]!=' ' && z[2]!='\t' ) return 0;
  return 2;
}

/*
** If *z points to horizontal rule markup, return the number of
** characters in that markup.  Otherwise return 0.
**
** Horizontal rule markup consists of four or more '-' characters
** at the beginning of a line followed by nothing but whitespace
** to the end of the line.
*/
static int is_horizontal_rule(const char *z){
  int i;
  for(i=0; z[i]=='-'; i++){}
  if( i<4 ) return 0;
  while( isspace(z[i]) && z[i]!='\n' ){ i++; }
  return z[i]=='\n' || z[i]==0 ? i : 0;
}

/*
** Return the number of characters in the URL that begins
** at *z.  Return 0 if *z is not the beginning of a URL.
**
** Algorithm: Advance to the first whitespace character or until
** then end of the string.  Then back up over the following
** characters:  .)]}?!"':;,
*/
static int is_url(const char *z){
  int i;
  int minlen = 6;
  switch( z[0] ){
    case 'h':
     if( strncmp(z,"http:",5)==0 ) minlen = 7;
     else if( strncmp(z,"https:",6)==0 ) minlen = 8;
     else return 0;
     break;
    case 'f':
     if( strncmp(z,"ftp://",6)==0 ) minlen = 7;
     else return 0;
     break;
    case 'm':
     if( strncmp(z,"mailto:",7)==0 ) minlen = 10;
     else return 0;
     break;
    default:
     return 0;
  }
  for(i=0; z[i] && !isspace(z[i]); i++){}
  while( i>0 ){
    switch( z[i-1] ){
      case '.':
      case ')':
      case ']':
      case '}':
      case '?':
      case '!':
      case '"':
      case '\'':
      case ':':
      case ';':
      case ',':
        i--;
        break;
      default:
        return i>=minlen ? i : 0;
    }
  }
  return 0;
}

/*
** Output N characters of text from zText.
*/
static void put_htmlized_text(const char **pzText, int N){
  if( N>0 ){
    char *z = htmlize(*pzText, N);
    cgi_printf("%s", z);
    free(z);
    *pzText += N;
  }
}

/*
** Search ahead in text z[] looking for a font terminator character
** "c".  The font terminator must be at the end of a word and it
** must occur before a paragraph break.  Also, z[] must begin a new
** word.  If any of these conditions are false, return false.  If
** all conditions are meet, return true.
*/
static int font_terminator(const char *z, int c){
  int seenNL = 0;
  if( isspace(*z) || *z==0 || *z==c ) return 0;
  z++;
  while( *z ){
    if( *z==c && !isspace(z[-1]) && is_eow(&z[1],0) ){
      return 1;
    }else if( *z=='\n' ){
      if( seenNL ) return 0;
      seenNL = 1;
    }else if( !isspace(*z) ){
      seenNL = 0;
    }
    z++;
  }
  return 0;
}

/*
** Output Wiki text while inserting the proper HTML control codes.
** The following formatting conventions are implemented:
**
**    *    Characters with special meaning to HTML are escaped.
**
**    *    Blank lines results in a paragraph break.
**
**    *    Paragraphs where the first line is indented by two or more
**         spaces are shown verbatim.  None of the following rules apply
**         to verbatim text.
**
**    *    Lines beginning with "*: " begin a bullet in a bullet list.
**
**    *    Lines beginning with "1: " begin an item in an enumerated list.
**
**    *    Words that begin and end with underscore (ex: _this_is_bold_)
**         are shown in bold.  Mid-word underscores are converted to
**         spaces.
**
**    *    Wiki pages names (Words in initial caps) are enclosed in an
**         appropriate hyperlink.
**
**    *    Words that begin with "http:" are enclosed in an appropriate
**         hyperlink.
**
**    *    Text of the form "#NNN" where NNN is a valid ticket number
**         is converted into a hyperlink to the corresponding ticket.
**
**    *    Text of the form "[NNN]" where NNN is a valid check-in number
**         becomes a hyperlink to the checkin.
*/
void output_wiki(const char *zText, const char *zLinkSuffix){
  int i, j;
  int inPRE = 0;
  int inUL = 0;
  int inOL = 0;
  int inB = 0;
  int inI = 0;
  int v;
  int wordStart = 1;     /* At the start of a word */
  int lineStart = 1;     /* At the start of a line */
  int paraStart = 1;     /* At the start of a paragraph */
  static int once = 1;
  static int nTicket, nCommit;
  if( once ){
    nTicket = atoi(db_short_query("SELECT max(tn) FROM ticket"));
    nCommit = atoi(db_short_query("SELECT max(cn) FROM chng"));
    once = 0;
  }

  i = 0;
  while( zText[i] ){
    char *z;
    int c = zText[i];

    if( paraStart ){
      put_htmlized_text(&zText, i);

      /* Blank lines at the beginning of a paragraph are ignored.
      */
      if( isspace(c) && (j = is_blank_line(&zText[i]))>0 ){
        zText += j;
        continue;
      }

      /* If the first line of a paragraph begins with a tab or with two
      ** or more spaces, then that paragraph is printed verbatim.
      */
      if( c=='\t' || (c==' ' && (zText[i+1]==' ' || zText[i+1]=='\t')) ){
        if( !inPRE ){
          if( inB ){ cgi_printf("</b>"); inB=0; }
          if( inI ){ cgi_printf("</i>"); inI=0; }
          if( inOL ){ cgi_printf("</ol>\n"); inOL=0; }
          if( inUL ){ cgi_printf("</ul>\n"); inUL=0; }
          cgi_printf("<p><pre>\n");
          inPRE = 1;
        }
      }
    }
    if( lineStart ){
      /* Blank lines in the middle of text cause a paragraph break
      */
      if( isspace(c) && (j = is_blank_line(&zText[i]))>0 ){
        put_htmlized_text(&zText, i);
        zText += j;
        if( inB ){ cgi_printf("</b>"); inB=0; }
        if( inI ){ cgi_printf("</i>"); inI=0; }
        if( inPRE ){ cgi_printf("</pre></p>\n"); inPRE = 0; }
        if( inOL && !is_enum(zText) ){ cgi_printf("</ol>\n"); inOL=0; }
        if( inUL && !is_bullet(zText) ){ cgi_printf("</ul>\n"); inUL=0; }
        if( zText[0]!=0 ){ cgi_printf("\n<p>"); }
        wordStart = lineStart = paraStart = 1;
        i = 0;
        continue;
      }
    }
    if( lineStart && !inPRE ){
      /* If we are not in verbatim text and a line begins with "*:", then
      ** generate a bullet.  Or if the line begins with "NNN:" where NNN
      ** is a number, generate an enumeration item.
      */
      if( (j = is_bullet(&zText[i]))>0 || (j = is_enum(&zText[i]))>0 ){
        put_htmlized_text(&zText, i);
        if( zText[0]=='*' ){
          if( inOL ) cgi_printf("</ol>\n");
          if( !inUL ) cgi_printf("<ul>\n");
          inUL = 1;
          inOL = 0;
        }else{
          if( inUL ) cgi_printf("</ul>\n");
          if( !inOL ) cgi_printf("<ol>\n");
          inUL = 0;
          inOL = 1;
        }
        zText += j;
        if( *zText ) zText++;
        i = 0;
        cgi_printf("<li>");
        wordStart = 1;
        lineStart = paraStart = 0;
        continue;
      }

      /* Four or more "-" characters on at the beginning of a line that
      ** contains no other text results in a horizontal rule.
      */
      if( c=='-' && (j = is_horizontal_rule(&zText[i]))>0 ){
        put_htmlized_text(&zText, i);
        if( inOL ){ cgi_printf("</ol>\n"); inOL=0; }
        if( inUL ){ cgi_printf("</ul>\n"); inUL=0; }
        cgi_printf("<hr>\n");
        zText += j;
        if( *zText ) zText++;
        i = 0;
        lineStart = wordStart = 1;
        paraStart = 0;
        continue;
      }
    }
    if( wordStart && !inPRE ){
      /* A wiki name at the beginning of a word which is not in verbatim
      ** text generates a hyperlink to that wiki page.
      ** 
      ** Special case: If the name is in CamelCase but ends with a "_", then
      ** suppress the "_" and do not generate the hyperlink.  This allows
      ** CamelCase words that are not wiki page names to appear in text.
      */
      if( g.okRdWiki && isupper(c) && (j = is_wiki_name_(&zText[i]))>0 ){
        put_htmlized_text(&zText, i);
        if( !isalpha(zText[j-1]) ){
          cgi_printf("%.*s", j-1, zText);
          if( zText[j-1]=='_' && inI ) j--;
          if( zText[j-1]=='*' && inB ) j--;
        }else{
          cgi_printf("<a href=\"wiki?p=%.*s%s\">%.*s</a>",
              j, zText, zLinkSuffix, j, zText);
        }
        zText += j;
        i = 0;
        wordStart = lineStart = paraStart = 0;
        continue;
      }

      /* A "_" at the beginning of a word puts us into an italic font.
      */
      if( c=='_' && !inB && !inI && font_terminator(&zText[i+1],c) ){
        put_htmlized_text(&zText, i);
        i = 0;
        zText++;
        cgi_printf("<i>");
        inI = 1;
        continue;
      }

      /* A "*" at the beginning of a word puts us into a bold font.
      */
      if( c=='*' && !inB && !inI && font_terminator(&zText[i+1],c) ){
        put_htmlized_text(&zText, i);
        i = 0;
        zText++;
        cgi_printf("<b>");
        inB = 1;
        continue;
      }


      /* Words that begin with "http:" or "https:" or "ftp:" or "mailto:"
      ** become hyperlinks.
      */
      if( (c=='h' || c=='f' || c=='m') && (j=is_url(&zText[i]))>0 ){
        put_htmlized_text(&zText, i);
        z = htmlize(zText, j);
        cgi_printf("<a href=\"%s\">%s</a>", z, z);
        free(z);
        zText += j;
        i = 0;
        wordStart = lineStart = paraStart = 0;
        continue;
      }

      /* If the user has read permission on tickets and a word is of the
      ** form "#NNN" where NNN is a sequence of digits, then generate a
      ** hyperlink to ticket number NNN.
      */
      if( c=='#' && g.okRead && (j = ndigit(&zText[i+1]))>0 
                 && is_eow(&zText[i+1+j],0)
                 && (v = atoi(&zText[i+1]))>0 && v<=nTicket ){
        put_htmlized_text(&zText, i);
        cgi_printf("<a href=\"tktview?tn=%d\">#%d</a>", v, v);
        zText += j;
        if( *zText ) zText++;
        i = 0;
        wordStart = lineStart = paraStart = 0;
        continue;
      }

      /* If the user has checkout permissions and a word is of the form
      ** "[NNN]" where NNN is a checkin number, then generate a hyperlink
      ** to check-in NNN.
      */
      if( c=='[' && g.okCheckout && (j = ndigit(&zText[i+1]))>0
                 && is_eow(&zText[i+j+2],0)
                 && (v = atoi(&zText[i+1]))>0 && v<=nCommit 
                 && zText[i+j+1]==']' ){
        put_htmlized_text(&zText, i);
        cgi_printf("<a href=\"chngview?cn=%d\">[%d]</a>", v, v);
        zText += j+1;
        if( *zText ) zText++;
        i  = 0;
        wordStart = lineStart = paraStart = 0;
        continue;
      }
    }

    /* A "*" or a "_" at the end of a word takes us out of bold or
    ** italic mode.
    */
    if( inB>0 && c=='*' && !isspace(zText[i-1]) && is_eow(&zText[i+1],0) ){
      inB = 0;
      put_htmlized_text(&zText, i);
      i = 0;
      zText++;
      cgi_printf("</b>");
      continue;
    }
    if( inI && c=='_' && !isspace(zText[i-1]) && is_eow(&zText[i+1],0) ){
      put_htmlized_text(&zText, i);
      i = 0;
      zText++;
      inI = 0;
      cgi_printf("</i>");
      continue;
    }
    if( wordStart ){
      wordStart = isspace(c) || c=='(' || c=='"';
    }else{
      wordStart = isspace(c);
    }
    lineStart = c=='\n';
    paraStart = 0;
    i++;
  }
  if( zText[0] ) cgi_printf("%h", zText);
  if( inB ) cgi_printf("</b>\n");
  if( inI ) cgi_printf("</i>\n");
  if( inOL ) cgi_printf("</ol>\n");
  if( inUL ) cgi_printf("</ul>\n");
  if( inPRE ) cgi_printf("</pre></p>\n");
}

/*
** Output text while inserting hyperlinks to ticket and checkin reports.
** Within the text, an occurance of "#NNN" (where N is a digit) results
** in a hyperlink to the page that shows that ticket.  Any occurance of
** [NNN] gives a hyperlink to check-in number NNN.
**
** (Added later:)  Also format the text as HTML.  Insert <p> in place
** of blank lines.  Insert <pre>..</pre> around paragraphs that are
** indented by two or more spaces.  Make lines that begin with "*:"
** or "1:" into <ul> or <ol> list elements.
**
** (Later:)  The formatting is now extended to include all of the
** Wiki formatting options.
*/
void output_formatted(const char *zText){
  output_wiki(zText,"");
}

/* 
** This routine alters a check-in message to make it more readable
** in a timeline.  The following changes are made:
**
** *:  Remove all leading whitespace.  This prevents the text from
**     being display verbatim.
**
** *:  If the message begins with "*:" or "N:" (where N is a number)
**     then convert the ":" into "." to disable the bullet or enumeration.
**
** *:  Change all newlines to spaces.  This will disable paragraph
**     breaks, verbatim paragraphs, enumerations, and bullet lists.
**
** *:  Collapse contiguous whitespace into a single space character
**
** *:  Truncate the string at the first whitespace character that
**     is more than mxChar characters from the beginning of the string.
**     Or if the string is longer than mxChar character and but there
**     was a paragraph break after mnChar characters, truncate at the
**     paragraph break.
**
** This routine changes the message in place.  It returns non-zero if
** the message was truncated and zero if the original text is still
** all there (though perhaps altered.)
*/
int output_trim_message(char *zMsg, int mnChar, int mxChar){
  int i, j, k;
  int brkpt = 0;    /* First paragraph break after zMsg[mnChar] */

  if( zMsg==0 ) return 0;
  for(i=0; isspace(zMsg[i]); i++){}
  if( zMsg[i]=='*' && (k = is_bullet(&zMsg[i]))>0 ){
    zMsg[i+1] = '.';
  }else if( isdigit(zMsg[i]) && (k = is_enum(&zMsg[i]))>0 ){
    zMsg[i+k-1] = '.';
  }
  for(j=0; zMsg[i]; i++){
    int c = zMsg[i];
    if( c=='\n' ){
      if( j>mnChar && is_blank_line(&zMsg[i+1]) && brkpt==0 ){
        brkpt = j;
      }
      c = ' ';
    }
    if( isspace(c) ){
      if( j>=mxChar ){
        zMsg[j] = 0;
        if( brkpt>0 ) zMsg[brkpt] = 0;
        return 1;
      }
      if( j>0 && !isspace(zMsg[j-1]) ){
        zMsg[j++] = ' ';
      }
    }else{
      zMsg[j++] = c;
    }
  }
  zMsg[j] = 0;
  return 0;
}

/*
** Append HTML text to the output that describes the formatting
** conventions implemented by the output_formatted() function
** above.
*/
void append_formatting_hints(void){
  @ <ul>
  @ <li><p>
  @ Blank lines result in a paragraph break.
  @ </p></li>
  @
  @ <li><p>
  @ If a paragraph is indented by a tab or by two or more spaces,
  @ it is displayed verbatim -- in a constant-width font with all
  @ spacing and line breaks preserved.
  @ </p></li>
  @
  @ <li><p>
  @ The characters "<tt>*:</tt>"  (asterisk-colon) at the very beginning
  @ of a line create a bullet in a bullet-list.
  @ The characters "<tt>1:</tt>"  (one-colon) at the very beginning
  @ of a line create a enumeration item in an enumerated list.
  @ </p></li>
  @
  @ <li><p>
  @ Words surrounded by underscores (ex: "<tt>_italic text_</tt>") are
  @ rendered in an italic font.  Words surrounded by asterisks
  @ (ex: "<tt>*bold text*</tt>") are rendered in a bold font.
  @ </p></li>
  @
  @ <li><p>
  if( g.okRead ){
    @ Text like "<tt>#123</tt>" becomes a hyperlink to ticket #123.
  }
  if( g.okCheckout ){
    @ Text like "<tt>[456]</tt>" is replaced by a hyperlink to
    @ check-in [456].
  }
  @ Text that looks like URL
  if( g.okRdWiki ){
    @ or a wiki page name 
  }
  @ becomes a hyperlink.
  @ </p></li>
  @
  @ <li><p>
  @ Four or more "-" characters on a line by themselves generate a
  @ horizontal rule (the &lt;hr&gt; markup of HTML).
  @ </p></li>
  if( g.okRdWiki ){
    @
    @ <li><p>
    @ Additional formatting information is available in the Wiki at
    @ <a href="wiki?p=FormattingWikiPages">FormattingWikiPages</a>.
    @ </p></li>
  }
  @ </ul>
}
